#include "ACO.h"
#include <iostream>
using namespace std;
#include <math.h>
#include <fstream>
using namespace std;
#include <sstream>
#include <time.h>

ACO::ACO(int nsize, vector<int> obs, int startpos, int endpos) 
{
  size = nsize;// size x and size y
  q0 = 0.5;
  t0 = 10; 
  numberOfAnts = 200;// number of ants
  E = 150;//maximum trasport number
  alpha = 2;//alphaspin->value();
  beta = 1;//betaspin->value();
  rho = 0.9;//rhospin->value();
  NumberOfGenerations = 100;//NumberOfGenerationsSpin->value();
  Q = 10*t0;//1000;//pherQuantity->value();
  
  startPosition = startpos;
  endPosition = endpos;//size*size-1;
  
  obstacles = obs;
}



vector<int> ACO::antACO(){
  finalSolution.clear();
  ofstream stream;
  stream.open( "ACOplan.txt",ios::app );
   if (stream.is_open())
    {
      time_t start;
      start = time (NULL);
      ACOSolution(size);
      optimize();
      time_t stop;
      stop = time (NULL);
	  stop = stop - start;				
	  stream<< "\n";
	  stream<<numberOfAnts<<" "<<NumberOfGenerations;
	  stream<<" "<<alpha<<" "<<beta<<" "<<E<<" ";
	  stream<< finalSolution.size();
	  stream<< " ";
	  stream<<stop<<" .end";
	}
   else
    	cout << "Unable to open file";
    return finalSolution;
}
bool ACO::IsDiagonal(int pos1,int pos2){
  if( (pos2 == pos1+size+1) && !IsObstacle(pos2-1) && !IsObstacle(pos1+1))
    return true;
  if( (pos2 == pos1-size-1) && !IsObstacle(pos1-1) && !IsObstacle(pos2+1))
    return true;
  if( (pos2 == pos1-size+1) && !IsObstacle(pos2-1) && !IsObstacle(pos1+1))
    return true;
  if( (pos2 == pos1+size-1) && !IsObstacle(pos2+1) && !IsObstacle(pos1-1))
    return true;
  return false;
}
void ACO::optimize()
{
  for(int i=0;i<(finalSolution.size()-2);i++){
    if(IsDiagonal(finalSolution[i], finalSolution[i+2]))
      finalSolution.erase(finalSolution.begin()+(i+1));
  }

  for(int j=0;j<(finalSolution.size()-2);j++){
    if(((finalSolution.at(j)%size) == (finalSolution.at(j+2)%size)) && 
        !IsObstacle(size*((int) finalSolution.at(j+1)/size) + finalSolution.at(j)%size) )
      finalSolution.at(j+1) = size*((int) finalSolution.at(j+1)/size) + finalSolution.at(j)%size;
    else if(((finalSolution.at(j)/size) == (finalSolution.at(j+2)/size)) &&
        !IsObstacle(size*((int) finalSolution.at(j)/size) + finalSolution.at(j+1)%size))
      finalSolution.at(j+1) = size*((int) finalSolution.at(j)/size) + finalSolution.at(j+1)%size;  
  } 
  for(int i=0;i<(finalSolution.size()-2);i++){
    if(IsDiagonal(finalSolution[i], finalSolution[i+2]))
      finalSolution.erase(finalSolution.begin()+(i+1));
  }
}
double ACO::heuristic(int currentpos)
{
  int x = currentpos%size;
  int y = currentpos/size;
  int temp = (int) sqrt(pow((endPosition%size-x),2)+pow((endPosition/size-y),2));
  return (double) 1/temp;
}
bool ACO::IsObstacle(int pos)
{
  for(int j=0;j<obstacles.size();j++)
  {
    if(pos == obstacles[j])
      return true;
  }
  return false;
}
bool ACO::InJourneyNodeList(int ant, int pos)
{
  try
  {
    if(ant==numberOfAnts){
      for(int i=0;i<solution.size();i++){
	    if(solution.at(i) == pos)
	      return true;
	    }
      }
    else{
      list<int>::iterator it;
      for ( it=journeynodeslists.at(ant).begin() ; it != journeynodeslists.at(ant).end(); it++ ){
        if(*it == pos)
          return true;
        }
      }	 
    } catch ( std::exception outOfRange ) // out_of_range exception
    {
	  cout << "\n\nException: " << outOfRange.what();
  }
  return false;
}
bool ACO::IsPermitted(int previousNode, int newNode){
  if((newNode==(previousNode+size)) || 
		  (newNode==(previousNode-size)) ||
		  (newNode==(previousNode+1)) ||
		  (newNode==(previousNode-1)))
    return false;
  else 
    return true;
}
int ACO::calculatePij(int nnode,vector<float> npher,int nk)
{
  double max = -1.0;
  double temp ;
  int node = nnode;
  int nextnode = -1;
  int previousnode = -1; 
  vector<float> pher(npher);
  int ant = nk;

  int bestNode = 0;
  double bestValue = -1;

  if(ant==numberOfAnts){
    int temp = solution.back();
    solution.pop_back();
    previousnode = solution.back();
    solution.push_back(temp);				
  }
  else{
	int temp = journeynodeslists.at(ant).back();
	journeynodeslists.at(ant).pop_back();
	previousnode = journeynodeslists.at(ant).back();
	journeynodeslists.at(ant).push_back(temp);
  }

  for(int i=-1;i<2;i++)
  {
    if(((node+i)<(size*size)) && !IsObstacle(node+i) && i!=0 && !InJourneyNodeList(ant,node+i) && 
		IsPermitted(previousnode,node+i))
    {
      if((node%size==0)){ 
        if(i>-1){
          double val = pow(pher[node+i],alpha)*pow(heuristic(node+i),beta);
          if(val>bestValue){
            bestNode = node+i;
            bestValue = val;				
          }
          temp=rand()*val/RAND_MAX;
          if(temp>max)
          {
            max=temp;
            nextnode = node+i;
          }
        }
      }
      else if((node+1)%size==0){
        if(i<1){
          double val = pow(pher[node+i],alpha)*pow(heuristic(node+i),beta);
          if(val>bestValue){
            bestNode = node+i;
            bestValue = val;				
          }
          temp=rand()*val/RAND_MAX;
          if(temp>max)
          {
            max=temp;
            nextnode = node+i;
          }
        }
      }
      else{
        double val = pow(pher[node+i],alpha)*pow(heuristic(node+i),beta);
        if(val>bestValue){
          bestNode = node+i;
          bestValue = val;				
        }
        temp=rand()*val/RAND_MAX;
        if(temp>max)
        {
          max=temp;
          nextnode = node+i;
        }
      }
    }
    if(((node-size+i)<(size*size)) &&!IsObstacle(node-size+i)&& !InJourneyNodeList(ant,node-size+i) && (node-size+i)>=0 
		&& IsPermitted(previousnode,node-size+i) && (!IsObstacle(node-size) & !IsObstacle(node+i)))
    {
      if((node%size==0)){ 
        if(i>-1){
          double val = pow(pher[node+i],alpha)*pow(heuristic(node+i),beta);
          if(val>bestValue){
            bestNode = node-size+i;
            bestValue = val;				
          }
          temp=rand()*val/RAND_MAX;
          if(temp>max)
          {
            max=temp;
            nextnode = node-size+i;
          }
        }
      }
      else if((node+1)%size==0){
        if(i<1){
          double val = pow(pher[node+i],alpha)*pow(heuristic(node+i),beta);
          if(val>bestValue){
            bestNode = node-size+i;
            bestValue = val;				
          }
          temp=rand()*val/RAND_MAX;
          if(temp>max)
          {
            max=temp;
            nextnode = node-size+i;
          }
        }
      }
      else{
        double val = pow(pher[node+i],alpha)*pow(heuristic(node+i),beta);
          if(val>bestValue){
            bestNode = node-size+i;
            bestValue = val;				
          }
          temp=rand()*val/RAND_MAX;
        if(temp>max)
        {
          max=temp;
          nextnode = node-size+i;
        }
      }
    }
    if(((node+size+i)<(size*size)) &&!IsObstacle(node+size+i)&& !InJourneyNodeList(ant,node+size+i) && (node+size+i)>=0 
		 && IsPermitted(previousnode,node+size+i) && (!IsObstacle(node+size) & !IsObstacle(node+i)))
    {
      if((node%size==0)){ 
        if(i>-1){
          double val = pow(pher[node+i],alpha)*pow(heuristic(node+i),beta);
          if(val>bestValue){
            bestNode = node+size+i;
            bestValue = val;				
          }
          temp=rand()*val/RAND_MAX;
          if(temp>max)
          {
            max=temp;
            nextnode = node+size+i;
          }
        }
      }
      else if((node+1)%size==0){
        if(i<1){
          double val = pow(pher[node+i],alpha)*pow(heuristic(node+i),beta);
          if(val>bestValue){
            bestNode = node+size+i;
            bestValue = val;				
          }
          temp=rand()*val/RAND_MAX;
          if(temp>max)
          {
            max=temp;
            nextnode = node+size+i;
          }
        }
      }
      else{
        double val = pow(pher[node+i],alpha)*pow(heuristic(node+i),beta);
        if(val>bestValue){
          bestNode = node+size+i;
          bestValue = val;				
        }
        temp=rand()*val/RAND_MAX;
        if(temp>max)
        {
          max=temp;
          nextnode = node+size+i;
        }
      }    
    }
  } 
  if((rand()/RAND_MAX)>q0)
	return bestNode;
  else
	return nextnode;
}
void ACO::ACOSolution(int nsize){
  vector<float> pher(size*size);
  vector<float> deltapher(size*size);
  vector<bool> greached(numberOfAnts);
  vector<bool> IsSuspended(numberOfAnts);
  vector<int> transportActions(numberOfAnts);

  for(int i=0;i<journeynodeslists.size();i++)
  {
	  journeynodeslists[i].clear(); 
    journeynodeslists[i].push_back(0); 
	}
  journeynodeslists.resize(numberOfAnts);
  int ng = 0;
  int finalSolutionCost = INT_MAX;
  
  
  //Step 1. Initialization
  srand ( time(NULL) );

    for(int i=0;i<(size*size);i++){
		pher[i] = t0;
		deltapher[i] = 0;
  }
  finalSolution.clear();
  overallbest.clear();
  //Step 2. starting tour
  while(ng<NumberOfGenerations)
  {
	for(int i=0;i<numberOfAnts;i++){
    greached[i] = false;
    IsSuspended[i] = false;
	transportActions[i]=0;
  }
  for(int i=0;i<journeynodeslists.size();i++)
  {
    journeynodeslists[i].clear(); 
    journeynodeslists[i].push_back(startPosition); 
  }
  //step 3. searching for neighborhood
  int ant = 0;
  for(ant=0;ant<numberOfAnts;ant++)
  {
	  int newnode = -1;
	
	  while((transportActions[ant]<E) )
	  {  
		
	    newnode = calculatePij(journeynodeslists[ant].back(),pher,
	                      ant); // calculate new node of ant ant
	    
	    if(newnode == -1){
			transportActions[ant]=E;
	      break;
	    }
		else
		{
			journeynodeslists[ant].push_back(newnode); //add the new node at the ant's trajetory
			transportActions[ant]++;//increase the ant's movements
		}
	    if(newnode == endPosition ){
	     greached[ant]=true;
		  transportActions[ant]=E;
		  break;
	    }        
	  }
  }
  
//  Step 4. Calculate the cost for each ant's journey

  vector<int> iterationbest;
  int iterationbsize = journeynodeslists[0].size();
  int overallbsize;
  if(overallbest.empty()){
	overallbsize = INT_MAX;//journeynodeslists[0].size();
	//overallbest.push_back(journeynodeslists[0]);
	}
  else
  overallbsize = overallbest[0].size();
  for(int ant=0;ant<numberOfAnts;ant++)
  {
	if(greached[ant]){
		if(journeynodeslists[ant].size()<iterationbsize){
			iterationbest.clear();
			iterationbest.push_back(ant);
			iterationbsize = journeynodeslists[ant].size();	
		}
		else if(journeynodeslists[ant].size()==iterationbsize){
			iterationbest.push_back(ant);
			iterationbsize = journeynodeslists[ant].size();
		}
		if(journeynodeslists[ant].size()<overallbsize){
			overallbest.clear();
			overallbest.push_back(journeynodeslists[ant]);
			overallbsize = journeynodeslists[ant].size();	
		}
		else if(journeynodeslists[ant].size()==iterationbsize){
			overallbest.push_back(journeynodeslists[ant]);
			overallbsize = journeynodeslists[ant].size();
		}
	}
  }
  for(int i=0;i<iterationbest.size();i++){
	int best = iterationbest[i];		
	for (list<int>::const_iterator di = journeynodeslists[best].begin(); di != journeynodeslists[best].end(); ++di)
					deltapher[*di] += Q/journeynodeslists[best].size();		
  }
  for(int i=0;i<overallbest.size();i++){		
	for (list<int>::const_iterator di = overallbest[i].begin(); di != overallbest[i].end(); ++di)
					deltapher[*di] += Q/overallbest[i].size();		
  }
  
  //Step 5. Update the pheromone factor
  for(int i =0 ;i<size*size;i++)
  {
    float phert = rho*pher[i] + deltapher[i];
    if(phert < (10*t0))
  	  pher[i] = phert;
  	deltapher[i] = 0;
  }
  solution.clear();
  solution.push_back(startPosition);
  bool goal = false;
	
//   find the final solution 

  if(!overallbest.empty()){
    if(overallbest[0].size()<finalSolutionCost)
    {
	  finalSolutionCost = overallbest[0].size();
	  finalSolution.clear();
	  for (list<int>::const_iterator di = overallbest[0].begin(); di != overallbest[0].end(); ++di)
		finalSolution.push_back(*di);
    }
  }
    ng++;
  }
  
}



