/* =========================================================================== 
*  Implementation of a CPG for crawling
*
*  Copyright (c) 2005 Ludovic Righetti <ludovic.righetti _at_ a3.epfl.ch>
*
*  May be copied or modified under the terms of the GNU General Public
*  License.  See COPYING for more information.
*
*  Contributors:
*  @CONTNAME@ <@CONTEMAIL@> 
*
*
*
*  First implemented: 20/1/06 Ludovic Righetti 
* 
*  History:
*  
=============================================================================*/

#ifndef CRAWLING_CPG_H
#define CRAWLING_CPG_H


//This is the CPG for crawling

#include <math.h>
#include <gsl/gsl_math.h>



#include <math/DynamicalSystem.h>
#include <iostream>
#include <fstream>
#include <ostream>

#define TROT 0
#define BOUND 1
#define PACE 2
#define WALK 3
#define NO_COUPLING 4

using namespace std;

class Crawling_CPG : public DynamicalSystem {

public:

  Crawling_CPG();
  virtual  ~Crawling_CPG();

  void setSensors(short sens[]);
  void setSensors(double sens[]);
  void setCurrentPosition(double pos[]);
  void setDiscreteGoals(double goals[2][4]);
  void setAmplitude(double amplitude);
  
  void enableFeedback(int en);
  void enableObserv(int en);


  void setParameters(Parameter* params);
  virtual void odeFunction(double t, const double y[], double dydt[]);

  int select_gait(int gait);

  void saveParameters(const char *name=NULL);

 private:
  
  double discrete_goals[2][4];
  double mu;
  double touch_sensors[4],current_position[8];
  int enable_feedback,enable_observ;
  int *current_gait;
  static int trot_matrix[4*4];
  static int walk_matrix[4*4];
  static int pace_matrix[4*4];
  static int bound_matrix[4*4];
  static int nocoupling_matrix[4*4];
  
  fstream paramFile;
};

#endif //CRAWLING_CPG_H
