/***************************************************************************
 *   Copyright (C) 2006 by Ludovic Righetti   *
 *   ludovic.righetti-at-epfl.ch   *
 *                                                                          *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "iCubControl.h"

iCubControl::iCubControl(int time_step)
{
  this->time_step = time_step; 
}


iCubControl::~iCubControl()
{
  target_file.close();
  real_file.close();
  sensor_file.close();
}


void iCubControl::set_positions( )
{
  int i;
  for (i= 0;i<NB_DOF;++i)
    servo_set_position(cub_servo[i],cub_target_pos[i]);
}

void iCubControl::get_positions()
{
  for(int i=0;i<NB_DOF;++i)
    {
      double tmp = servo_get_position(cub_servo[i]);
      cub_vel[i] = (tmp - cub_pos[i])/((float)(time_step)/1000.0);
      cub_pos[i] = tmp;
    }
  
}

void iCubControl::get_sensor_values()
{
  for(int i=0;i<NB_TOUCH;i++)
    cub_sensor_values[i] = touch_sensor_get_value(touch_sensors[i]);
}

void iCubControl::print_positions( )
{
  int i;
  for (i = 0; i<25 ;++i)
    cout << "servo " <<i << " " << cub_target_pos[i] << " ";
  cout << endl;
}

void iCubControl::reset( )
{
  int i;  
  
  cub_servo[0] = robot_get_device("NECK_TILT");
  cub_servo[1] = robot_get_device("left_leg_1");
  cub_servo[2] = robot_get_device("left_leg_2");
  cub_servo[3] = robot_get_device("left_leg_3");
  cub_servo[4] = robot_get_device("left_knee");
  cub_servo[5] = robot_get_device("right_leg_1");
  cub_servo[6] = robot_get_device("right_leg_2");
  cub_servo[7] = robot_get_device("right_leg_3");
  cub_servo[8] = robot_get_device("right_knee");
  cub_servo[9] = robot_get_device("left_arm_1");
  cub_servo[10] = robot_get_device("left_arm_2");
  cub_servo[11] = robot_get_device("left_arm_3");
  cub_servo[12] = robot_get_device("left_elbow");
  cub_servo[13] = robot_get_device("right_arm_1");
  cub_servo[14] = robot_get_device("right_arm_2");
  cub_servo[15] = robot_get_device("right_arm_3");
  cub_servo[16] = robot_get_device("right_elbow");
  cub_servo[17] = robot_get_device("torso_1");
  cub_servo[18] = robot_get_device("torso_2");
  cub_servo[19] = robot_get_device("torso_3");
  cub_servo[20] = robot_get_device("NECK_PAN");
  cub_servo[21] = robot_get_device("right_ankle");
  cub_servo[22] = robot_get_device("left_ankle");
  cub_servo[23] = robot_get_device("left_wrist_1");
  cub_servo[24] = robot_get_device("left_wrist_2");
  cub_servo[25] = robot_get_device("right_wrist_1");
  cub_servo[26] = robot_get_device("right_wrist_2");
  cub_servo[27] = robot_get_device("left_forearm");
  cub_servo[28] = robot_get_device("right_forearm");
  
  gps = robot_get_device("gps");
  gps_enable(gps,1);

  touch_sensors[0] = robot_get_device("left_hand_touch");
  touch_sensors[1] = robot_get_device("right_hand_touch");
  touch_sensors[2] = robot_get_device("left_knee_touch");
  touch_sensors[3] = robot_get_device("right_knee_touch");

 receiver = robot_get_device("receiver");
 receiver_enable(receiver,1);  
 if (receiver==0) robot_console_printf("missing receiver\n");

  for(i=0;i<NB_TOUCH;i++)
    {
      if(!touch_sensors[i])
	cout << "error for sensor " << i << endl;
      else
	touch_sensor_enable(touch_sensors[i],time_step);
    }

  for(i=0;i<NB_DOF;++i)
    {
      if(!cub_servo[i])
	cout << "error for servo " << i << endl;
      else
	{
	  servo_enable_position(cub_servo[i],time_step);
	}
      cub_target_pos[i] = 0.0;
      cub_pos[i] = 0.0;
    }
  
  return;
}

void iCubControl::saveRealTrajectory(const char *name)
{

  if(!real_file.is_open())
    {
      if(name==NULL)
	real_file.open("real_trajectory.dat",fstream::out);
      else
	real_file.open(name,fstream::out);
    }
  else
    if(name!=NULL)
      {
	real_file.close();
	real_file.open(name,fstream::out);
      }
  
  for(int i =0;i<NB_DOF;i++)
    {
      real_file << cub_pos[i] << " ";
    }
  real_file << endl;
  //real_file.flush();
}

void iCubControl::savePlannedTrajectory(const char *name)
{
  if(!target_file.is_open())
    {
      if(name==NULL)
	target_file.open("target_trajectory.dat",fstream::out);
      else
	target_file.open(name,fstream::out);
    }
  else
    if(name!=NULL)
      {
	target_file.close();
	target_file.open(name,fstream::out);
      }
  
  //record planned trajectory
  for(int i =0;i<NB_DOF;i++)
    {
      target_file << cub_target_pos[i] << " ";
    }
  target_file << endl;
  
  //target_file.flush();
}


void iCubControl::saveSensorValues(const char *name)
{
  if(!sensor_file.is_open())
    {
      if(name==NULL)
	sensor_file.open("sensor_values.dat",fstream::out);
      else
	sensor_file.open(name,fstream::out);
    }
  else
    if(name!=NULL)
      {
	sensor_file.close();
	sensor_file.open(name,fstream::out);
      }

  for(int i=0;i<NB_TOUCH;i++)
    sensor_file << cub_sensor_values[i] << " ";

  sensor_file << endl;

  //sensor_file.flush();
}

void iCubControl::savePosition(const char *name)
{
  if(!position_file.is_open())
    {
      if(name==NULL)
	position_file.open("robot_position.dat",fstream::out);
      else
	position_file.open(name,fstream::out);
    }
  else
    if(name!=NULL)
      {
	position_file.close();
	position_file.open(name,fstream::out);
      }

  for(int i=0;i<3;i++)
    position_file << robot_current_position[i] << " ";
  
  position_file << endl;
}

void iCubControl::supervisorGetiCubPosition()
{
  ///Get the robot and servo nodes///
  robot = supervisor_node_get_from_def("iCub");
  supervisor_field_get(robot,SUPERVISOR_FIELD_TRANSLATION_AND_ROTATION,
		       (void *)iCub_position,1);
  
}


void iCubControl::supervisorGetRobotPosition()
{
  ///Get the robot and servo nodes///
  robot = supervisor_node_get_from_def("iCub");
  supervisor_field_get(robot,SUPERVISOR_FIELD_TRANSLATION_AND_ROTATION,
		       (void *)robot_current_position,128);
  
  for(int l=0;l<NB_DOF;l++) 
    {
      servo[l]=supervisor_node_get_from_def(servo_name[l]);
      supervisor_field_get(servo[l],SUPERVISOR_FIELD_TRANSLATION_AND_ROTATION,
			   (void *)servo_current_position[l],128);
    }
  
  robot_step(128); /* get and save initial positions of robots and servos */

  for(int m=0;m<7;m++)
    robot_initial_position[m]=robot_current_position[m];

  // we have to correct the rotation angle which may have changed due
  // to physics simulation before the supervisor process starts
  robot_initial_position[3]=1;
  robot_initial_position[4]=0;
  robot_initial_position[5]=0;

  for(int l=0;l<NB_DOF;l++) 
    {
      if (!supervisor_node_was_found(servo[l]))
	cout << "SUPERVISOR: servo " << servo_name[l] << " not found" << endl;
      else
	for(int m=0;m<7;m++)
	  servo_initial_position[l][m]=servo_current_position[l][m];
    }
  
  this->get_positions();
  for(int i=0;i<NB_DOF;i++)
    cub_initial_position[i] = cub_pos[i];
}


void iCubControl::supervisorSetRobotPosition()
{
  for(int i = 0;i<NB_DOF;i++)
    cub_target_pos[i] = cub_initial_position[i];
  
  this->set_positions();
  robot_step(1000);
  
  supervisor_field_set(robot,SUPERVISOR_FIELD_TRANSLATION_AND_ROTATION,
		       robot_initial_position);
  
  for(int l=0;l<NB_DOF;l++)
    supervisor_field_set(servo[l],SUPERVISOR_FIELD_TRANSLATION_AND_ROTATION,
			 servo_initial_position[l]);

  supervisor_simulation_physics_reset();

  robot_step(500);
}




