/***************************************************************************
 *   Copyright (C) 2006 by Ludovic Righetti   *
 *   ludovic.righetti-at-epfl.ch   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef ICUBCONTROL_H
#define ICUBCONTROL_H

#include <device/robot.h>
#include <device/servo.h>
#include <device/touch_sensor.h>
#include <device/custom_robot.h>
#include <device/gps.h>
#include <device/supervisor.h>
#include <device/receiver.h>
#include <iostream>
#include <fstream>
#include <ostream>
#include <cstdlib>
#include <math.h>
#include <string.h>


#include <gsl/gsl_sf.h>
#include <gsl/gsl_math.h>

#define left_leg_1 1
#define left_leg_2 2
#define left_leg_3 3
#define left_knee 4

#define right_leg_1 5
#define right_leg_2 6
#define right_leg_3 7
#define right_knee 8

#define left_arm_1 9
#define left_arm_2 10
#define left_arm_3 11
#define left_elbow 12

#define right_arm_1 13
#define right_arm_2 14
#define right_arm_3 15
#define right_elbow 16

#define neck_tilt 0
#define torso_1 17
#define torso_2 18
#define torso_3 19
#define neck_pan 20

#define right_ankle 21
#define left_ankle 22

#define left_wrist_1 23
#define left_wrist_2 24
#define right_wrist_1 25
#define right_wrist_2 26

#define left_forearm 27
#define right_forearm 28

#define NB_DOF 29

#define NB_TOUCH 4
#define left_hand_touch 0
#define right_hand_touch 1
#define left_knee_touch 2
#define right_knee_touch 3


/**
	@author Ludovic Righetti <ludovic.righetti-at-epfl.ch>
*/



using namespace std;

static char   *servo_name[NB_DOF]=
{
    "NECK_TILT","NECK_PAN",
    "left_leg_1","left_leg_2","left_leg_3","left_knee","left_ankle",
    "right_leg_1","right_leg_2","right_leg_3","right_knee","right_ankle",
    "left_arm_1","left_arm_2","left_arm_3","left_elbow",
    "left_forearm","left_wrist_1","left_wrist_2",
    "right_arm_1","right_arm_2","right_arm_3","right_elbow",
    "right_forearm","right_wrist_1","right_wrist_2",
    "torso_1","torso_2","torso_3"
  };

class iCubControl{
 public:
  iCubControl(int time_step=1);
  
  virtual ~iCubControl();
  
  void set_positions();
  void get_positions();
  void get_sensor_values();

  void print_positions();
  
  void saveRealTrajectory(const char *name=NULL);
  void savePlannedTrajectory(const char *name=NULL);
  void saveSensorValues(const char *name=NULL);
  void savePosition(const char *name=NULL);

  void supervisorGetRobotPosition();
  void supervisorSetRobotPosition();
  void supervisorGetiCubPosition();

  virtual void reset();  
    DeviceTag gps;

  virtual int run(int dt) = 0;
  
 protected:
  DeviceTag cub_servo[NB_DOF];
  DeviceTag touch_sensors[NB_TOUCH];
  float cub_pos[NB_DOF];
  float cub_vel[NB_DOF];
  float cub_target_pos[NB_DOF];
  short cub_sensor_values[NB_TOUCH];

  int time_step;

  fstream target_file,real_file,sensor_file,position_file;
  
  ////supervisor stuff////
  float robot_initial_position[7];
  float robot_current_position[7];
  float servo_initial_position[NB_DOF][7];
  float servo_current_position[NB_DOF][7];
  float iCub_position[7];

  float cub_initial_position[NB_DOF];
  
  NodeRef robot;
  NodeRef servo[NB_DOF];
  NodeRef mark;
    
  DeviceTag receiver;
};

#endif
