#include "iCubCrawl.h"

#define X_AXIS 1
#define Z_AXIS 2
#define numberOfObstacles 5

int emph=1;
int revert=0;
int first=1;


iCubCrawl::iCubCrawl(int time_step) : iCubControl(time_step)
{
  mu_crawl = 1.0;
  mu_reach = -5.0;
  loop_time =0.0;

  elbow=0;
  state=CRAWL;
  time_collision=1;
  initial=0;

  feedback=0;
  observer=0;
  gait=TROT;

}

iCubCrawl::~iCubCrawl()
{
  stateFile.close();
  markFile.close();
}

void iCubCrawl::reset()
{
  iCubControl::reset();
  this->init(1.0,0.6);
}  



void iCubCrawl::saveState(const char *name)
{
 if(!stateFile.is_open())
    {
      if(name==NULL)
	stateFile.open("CPGstate.dat",fstream::out);
      else
	stateFile.open(name,fstream::out);
    }
  else
    if(name!=NULL)
      {
	stateFile.close();
	stateFile.open(name,fstream::out);
      }

  for(int i=0;i<CPGcontroller.getDimension();i++)
    stateFile << CPGcontroller.getState(i)->getValue() << " ";
  
  stateFile << endl;
}

void iCubCrawl::init(double om_sw,double om_st)
{
  robot_console_printf("Start\n");
  size = 8;
  goalNode = size*size-1;
  obstacles.clear();
  for(int i=0;i<numberOfObstacles;i++)   //numberOfObstacles
  {
    std::string def="OBSTACLE_";
    std::string s;
    std::stringstream out;
    out << i+1;
    s = out.str();
    def.append(s);
    char *buf = new char[strlen(def.c_str())];
    strcpy(buf,def.c_str());
    obstacle = supervisor_node_get_from_def(buf);
    supervisor_field_get (obstacle,
                          SUPERVISOR_FIELD_TRANSLATION_AND_ROTATION,
                          (void *)ob_translation_and_rotation[i], time_step );
    robot_step(0);
    for(int j=0;j<7;j++)
      initial_ob_translation_and_rotation[i][j] = ob_translation_and_rotation[i][j];
    int tempnode;
    tempnode = ((size-(int)ob_translation_and_rotation[i][2])-1)*size+
            (int)(ob_translation_and_rotation[i][0]);
    obstacles.push_back(tempnode);
    delete []buf;
  }
  robot_console_printf("obstacles found");
  ACO *aco = new ACO(size, obstacles, 0, size*size-1);
  path = aco->antACO();
  for(int i=0;i<path.size();i++)
    robot_console_printf("%d\n",path.at(i));
  currNode=0;
  previousNode=-1;
  pathCounter=1;
  previousOrientation=0.0;
  currOrientation = "N";
  turn_right=false;
  turn_left=false;
  count1=0;
  time=0;
  
  newOrientation = findOrientation(currNode,path[pathCounter]);
  left_turn = 1.0;
  right_turn = 1.0;


  //char file_name;
  markFile.open("markPosition.dat",fstream::in);

 
  ///Webots init stuff
  robot_step(2*time_step);
  this->supervisorGetRobotPosition();
  get_positions();
  
  robot_keyboard_enable( 25*time_step);
  
 
  for(int i =0;i<NB_DOF;i++)
    cub_target_pos[i] = cub_pos[i];

  
  //CPG init
  Parameter myParam[5];
  myParam[1] = om_sw * 2 * M_PI; //omega_swing
  myParam[2] = om_st * 2 * M_PI; //omega stance
  myParam[3] = 5.0; //alpha
  myParam[4] = 20.0; //gamma
  
  CPGcontroller.setParameters(myParam);

  CPGcontroller.setAmplitude(mu_crawl);

  arm_setpos = 1.57;
  leg_setpos = 1.3;
  arm_amp = 0.3;
  leg_amp = 0.3;
  
  double init[28];
  init[0] = 0.0;//((double)cub_pos[left_arm_1] - arm_setpos)/arm_amp;
  init[1] = 1.0;//sqrt(fabs(1.0-init[0]*init[0]));
  init[2] = 0.0;////((double)cub_pos[right_arm_1] - arm_setpos)/arm_amp;
  init[3] = -1.0;//sqrt(fabs(1.0-init[2]*init[2]));
  init[4] = 0.0;//((double)cub_pos[left_leg_1] - leg_setpos)/leg_amp;
  init[5] = -1.0;//sqrt(fabs(1.0-init[4]*init[4]));
  init[6] = 0.0;//((double)cub_pos[right_leg_1] - leg_setpos)/leg_amp;
  init[7] = 1.0;//sqrt(fabs(1.0-init[6]*init[6]));

  init[8] = myParam[1];
  init[9] = myParam[1];
  init[10] = myParam[1];
  init[11] = myParam[1];
  
  for(int i=12; i<28; i++){
    init[i]=0.0;} //discrete system

  loop_time=robot_get_time();

  CPGcontroller.setStates(init);
  CPGcontroller.enableObserv(observer);
  CPGcontroller.select_gait(gait);

}
double iCubCrawl::findOrientation(int currNode, int nextNode)
{
  const float *matrix = gps_get_matrix(gps);
  float euler[3];
  gps_euler(matrix,euler);
  int diff =  nextNode - currNode ;
  if(currOrientation == "N" || currOrientation == "S" )
  {
    if(diff == size){            
      previousOrientation=0.0;
      turn_left=false;
      turn_right=false;
      currOrientation = "N";
      return 0.0;
    }
    if(diff == (size-1)){
      previousOrientation=0.8;
      turn_left=true;
      turn_right=false;
      currOrientation = "NW";
      return 0.8;
      }
    if(diff == (size+1))
    {
      previousOrientation=-0.8;
      turn_left=false;
      turn_right=true;
      currOrientation = "NE";
      return -0.8;
      }
    if(diff == -size)
    {
      previousOrientation=0.0;
      turn_left=false;
      turn_right=false;
      currOrientation = "S";
      return 0.0;
      }
    if(diff == -(size-1))
    {
      previousOrientation=-0.8;
      turn_left=true;
      turn_right=false;
      currOrientation = "SE";
      return -0.8;
     }
     if(diff == -(size+1))
     {
       previousOrientation=0.8;
       turn_left=false;
       turn_right=true;
       currOrientation = "SW";
       return 0.8;
       }
      if(diff == 1)
      {
        if(currOrientation == "N"){
         previousOrientation=-1.6;
         turn_left=false;
         turn_right=true;
         currOrientation = "E";
         return -1.6;
         }
        if(currOrientation == "S"){
         previousOrientation=-1.6;
         turn_left=true;
         turn_right=false;
         currOrientation = "E";
         return -1.6;
         }
      }
      if(diff == -1)
      {
        if(currOrientation == "N"){
         previousOrientation=1.6;
         turn_left=true;
         turn_right=false;
         currOrientation = "W";
         return 1.6;
         }
        if(currOrientation == "S"){
         previousOrientation=1.6;
         turn_left=false;
         turn_right=true;
         currOrientation = "W";
         return 1.6;
         }
      }
      return previousOrientation;
  }
  else if(currOrientation == "E")
  {
    if(diff == 1)
    {
      previousOrientation=-1.6;
      turn_left=false;
      turn_right=false;
      currOrientation = "E";
      return -1.6;
      }
    if(diff == (size+1))
    {
      previousOrientation=-0.8;
      turn_left=true;
      turn_right=false;
      currOrientation = "NE";
      return -0.8;
      }
    if(diff == -(size-1))
    {
      previousOrientation=-0.8;
      turn_left=false;
      turn_right=true;
      currOrientation = "SE";
      return -0.8;
      }
    if(diff == size)
    {
      previousOrientation=0.0;
      turn_left=true;
      turn_right=false;
      currOrientation = "N";
      return 0.0;
      }
    if(diff == -size)
    {
      previousOrientation=0.0;
      turn_left=false;
      turn_right=true;
      currOrientation = "S";
      return 0.0;
      }
    return previousOrientation;
  }
  else if(currOrientation == "W")
  {
    if(diff == -1)
    {
      previousOrientation=1.6;
      turn_left=false;
      turn_right=false;
      currOrientation = "W";
      return 1.6;
      }
    if(diff == (size+1))
    {
      previousOrientation=0.8;
      turn_left=false;
      turn_right=true;
      currOrientation = "NW";
      return 0.8;
      }
    if(diff == -(size+1))
    {
      previousOrientation=0.8;
      turn_left=true;
      turn_right=false;
      currOrientation = "SW";
      return 0.8;
      }
    if(diff == size)
    {
      previousOrientation=0.0;
      turn_left=false;
      turn_right=true;
      currOrientation = "N";
      return 0.0;
      }
    if(diff == -size)
    {
      previousOrientation=0.0;
      turn_left=true;
      turn_right=false;
      currOrientation = "S";
      return 0.0;
      }
    return previousOrientation;
  }
  else if(currOrientation == "NE" || currOrientation == "SE")//euler[1]>-1.1 && euler[1]<-0.35)
  {
    if(diff == (size+1))
    {
      previousOrientation=-0.8;
      turn_left=false;
      turn_right=false;
      currOrientation = "NE";
      return -0.8;
      }
    if(diff == size)
    {
      previousOrientation=0.0;
      turn_left=true;
      turn_right=false;
      currOrientation = "N";
      return 0.0;
      }
    if(diff == 1)
    {
      if(euler[0]<0.0){
        previousOrientation=-1.6;
        turn_left=false;
        turn_right=true;
        currOrientation = "E";
        return -1.6;
       }
      else if(euler[0]>0.0){
        previousOrientation=-1.6;
        turn_left=true;
        turn_right=false;
        currOrientation = "E";
        return -1.6;
        }
      }
      if(diff == -(size-1))
      {
        previousOrientation=-0.8;
        turn_left=false;
        turn_right=false;
        currOrientation = "SE";
        return -0.8;
        }
      if(diff == -size)
      {
        previousOrientation=0.0;
        turn_left=false;
        turn_right=true;
        currOrientation = "S";
        return 0.0;
        }
      return previousOrientation;
      }
  else if(currOrientation == "NW" || currOrientation == "SW")//euler[1]>0.35 && euler[1]<1.1)
  {
    if(diff == (size-1))
    {
      previousOrientation=0.8;
      turn_left=false;
      turn_right=false;
      currOrientation = "NW";
      return 0.8;
      }
    if(diff == size)
    {
      previousOrientation=0.0;
      turn_left=false;
      turn_right=true;
      currOrientation = "N";
      return 0.0;
      }
    if(diff == -1)
    {
      if(euler[0]<0.0){
        previousOrientation=-1.6;
        turn_left=true;
        turn_right=false;
        currOrientation = "W";
        return -1.6;
       }
	  else if(euler[0]>0.0){
        previousOrientation=-1.6;
        turn_left=false;
        turn_right=true;
        currOrientation = "W";
        return -1.6;
        }
      }
      if(diff == -(size+1))
      {
        previousOrientation=-0.8;
        turn_left=false;
        turn_right=false;
        currOrientation = "SW";
        return -0.8;
        }
      if(diff == -size)
      {
        previousOrientation=0.0;
        turn_left=true;
        turn_right=false;
        currOrientation = "S";
        return 0.0;
      }
      return previousOrientation;
  }
  return previousOrientation;
}

 bool iCubCrawl::IsObstacle(int pos)
 {
  for(int j=0;j<obstacles.size();j++)
  {
    if(pos == obstacles[j])
      return true;
  }
  return false;
 }
 bool iCubCrawl::IsPlanChanged()
 {
    for(int i=0;i<numberOfObstacles;i++)
    {
      for(int j=0;j<7;j++)
      {
        if( initial_ob_translation_and_rotation[i][j] != ob_translation_and_rotation[i][j])
          return true;
      }    
    }
    return false;
 }

int iCubCrawl::run(int dt)
{
  time++;
  const float *matrix = gps_get_matrix(gps);
  float euler[3];
  gps_euler(matrix,euler);
  int key = robot_keyboard_get_key();
  if(key) {
   
    switch (key){
      case 'P': 
                supervisorGetiCubPosition();
                get_sensor_values();
                robot_console_printf("2d-ground position: [%g %g]\n",
                    iCub_position[0],iCub_position[2]);
                robot_console_printf("GPS positions: %.3f , %.3f , %.3f \n Orientation: % .3f\n",
                    gps_position_x(matrix),gps_position_y(matrix),
                    gps_position_z(matrix),euler[1]);
                robot_console_printf("GPS positions: %.0f , %.0f , %.0f \n Orientation: % .3f\n % .3f, % .3f",
                    gps_position_x(matrix),gps_position_y(matrix),
                    gps_position_z(matrix),euler[1],euler[0],euler[2]);
                robot_console_printf("torso_3: %g\n",cub_target_pos[19]);
                robot_console_printf("torso_2: %g\n",cub_target_pos[18]);
                robot_console_printf("right turn: %g\n left turn: %g\n",right_turn,left_turn);
                break;
      case 'M':
                for(int i=0;i<16;i++)
                  robot_console_printf("matrix[%d]: %g\n",i,matrix[i]);
                break;
      case 'S': 
                if(cub_target_pos[torso_3]<0.43){
                  cub_target_pos[torso_3]+= 0.05;
                  cub_target_pos[torso_2]+= 0.012;
                  left_turn+= 0.025;
                  right_turn+= -0.025;
                }
                break;
      case 'A': 
                if(cub_target_pos[torso_3]>-0.43){
                  cub_target_pos[torso_3]+= -0.05;
                  cub_target_pos[torso_2]+= -0.012;
                  left_turn+= -0.025;
                  right_turn+= 0.025;
                }
                break;
      case 'X':   
                cub_target_pos[torso_3]= 0.0;
                cub_target_pos[torso_2]= 0.0;
                left_turn=1.0;
                right_turn=1.0;
                break;  
      case 'O': 
                robot_console_printf("currNode: %d\n",currNode);
                robot_console_printf("newNode: %d\n",path[pathCounter]);
                robot_console_printf("newOrientation: %g\n",newOrientation);
                break;
    }
  }

  if(IsPlanChanged()){
    robot_console_printf("Changes detected at the obstacles' positions...");
    obstacles.clear();
    for(int i=0;i<numberOfObstacles;i++)
    {
      int newob;
      newob = ((size-(int)ob_translation_and_rotation[i][2])-1)*size+
              (int)(ob_translation_and_rotation[i][0]);
      obstacles.push_back(newob); 
    }
    robot_console_printf("Replanning...");
    ACO *aco = new ACO(size, obstacles, currNode, size*size-1);
    path = aco->antACO();
    for(int i=0;i<path.size();i++)
      robot_console_printf("%d\n",path.at(i));
    pathCounter=1;
    for(int i=0;i<numberOfObstacles;i++)
      for(int j=0;j<7;j++)
        initial_ob_translation_and_rotation[i][j] = ob_translation_and_rotation[i][j];
  }
  
  //we get values from simulation - sensors and positions
  get_positions();
  int currX = (int) gps_position_x(matrix);
  int currZ = (int) gps_position_z(matrix);
  int tempnode;
  tempnode = (size-(int)(gps_position_z(matrix))-1)*size+
            (int)(gps_position_x(matrix));
  if((gps_position_z(matrix)< ((int)(gps_position_z(matrix))+0.75)) &&
      (gps_position_z(matrix)> ((int)(gps_position_z(matrix))+0.25)) &&
      (gps_position_x(matrix)< ((int)(gps_position_x(matrix))+0.75)) &&
      (gps_position_x(matrix)> ((int)(gps_position_x(matrix))+0.25))){
        previousNode=currNode;
        currNode = tempnode;
      }
  int newNode = path[pathCounter];
  if(currNode==goalNode)
    return time_step;
  if (currNode==newNode){
      pathCounter++;
      newNode = path[pathCounter];
      newOrientation = findOrientation(currNode,newNode);
      robot_console_printf("previousNode: %d,currNode: %d, newNode: %d, pathCounter: %d\n",previousNode,currNode,newNode,pathCounter);             
  }
      if(time%10){

        float X_Goal = newNode%size + 0.5;
        float Z_Goal = size - (int)(newNode/size) - 0.5;
        float tempOrientation = atan((X_Goal - gps_position_x(matrix))/
                                  (Z_Goal - gps_position_z(matrix)));
        
        
        if((currOrientation=="E" && euler[0]>0.3 && euler[0]<2.1 && euler[1]>-1.35) ||       //left
            (currOrientation=="E" && gps_position_z(matrix)>(Z_Goal+0.2)) ||
            (currOrientation=="W" && euler[0]<-0.3 && euler[1]<1.35) ||
            (currOrientation=="W" && gps_position_z(matrix)<(Z_Goal-0.2))) 
            {
              if(cub_target_pos[torso_3]>-0.43){
                cub_target_pos[torso_3]+= -0.01;
                cub_target_pos[torso_2]+= -0.0025;
                left_turn+= -0.003;
                right_turn+= 0.003;
                }
            }
        else if((currOrientation=="E" && euler[0]<-0.3) && euler[1]>-1.35 ||           //right
            (currOrientation=="E" && gps_position_z(matrix)<(Z_Goal-0.2)) ||
            (currOrientation=="W" && euler[0]>0.3) && euler[1]<1.35 ||
            (currOrientation=="W" && gps_position_z(matrix)>(Z_Goal+0.2)))
            {
              if(cub_target_pos[torso_3]<0.43){
                cub_target_pos[torso_3]+= 0.01;
                cub_target_pos[torso_2]+= 0.0025;
                left_turn+= 0.003;
                right_turn+= -0.003;
                }
            }
        else if((currOrientation=="NE" && euler[1]<(tempOrientation-0.1)) ||   //left
                (currOrientation=="NW" && euler[1]<(tempOrientation-0.1))||
                (currOrientation=="SE" && euler[1]>(tempOrientation+0.1))||
                (currOrientation=="SW" && euler[1]>(tempOrientation+0.1)) ||
                (currOrientation=="N" && euler[1]<(tempOrientation-0.1)) ||
                (currOrientation=="N" && gps_position_x(matrix)>(X_Goal+0.25)) ||
                (currOrientation=="S" && euler[1]>(tempOrientation+0.1)) ||
                (currOrientation=="S" && gps_position_x(matrix)<(X_Goal-0.25))){
                  newOrientation = tempOrientation;
                  if(cub_target_pos[torso_3]>-0.43){
                    cub_target_pos[torso_3]+= -0.01;
                    cub_target_pos[torso_2]+= -0.0025;
                    left_turn+= -0.003;
                    right_turn+= 0.003;
                    }
                }
      else if((currOrientation=="NE" && euler[1]>(tempOrientation+0.1)) ||   //right
              (currOrientation=="NW" && euler[1]>(tempOrientation+0.1))||
              (currOrientation=="SE" && euler[1]<(tempOrientation-0.1))||
              (currOrientation=="SW" && euler[1]<(tempOrientation-0.1)) || 
              (currOrientation=="N" && euler[1]>(tempOrientation+0.1)) ||
              (currOrientation=="N" && gps_position_x(matrix)<(X_Goal-0.25)) ||
              (currOrientation=="S" && euler[1]<(tempOrientation-0.1)) ||
              (currOrientation=="S" && gps_position_x(matrix)>(X_Goal+0.25))){
                  newOrientation = tempOrientation;
                  if(cub_target_pos[torso_3]<0.43){
                    cub_target_pos[torso_3]+= 0.01;
                    cub_target_pos[torso_2]+= 0.0025;
                    left_turn+= 0.003;
                    right_turn+= -0.003;
                    }
                }
      else if(fabs(euler[1])>(fabs(newOrientation)-0.05) && fabs(euler[1])<(fabs(newOrientation)+0.05)){
        count1++;
        if(count1>10){
          cub_target_pos[torso_3]= 0.0;
          cub_target_pos[torso_2]= 0.0;
          left_turn=1.0;
          right_turn=1.0;
          turn_left=turn_right=false;
          count1=0;
          }
      }
  }
  
  ///for the pos, we scale the position for the CPG
  double thepos[8];
  thepos[0] = ((double)cub_pos[left_arm_1] - arm_setpos)/arm_amp;
  thepos[1] = (double)cub_vel[left_arm_1]/arm_amp;
  thepos[2] = ((double)cub_pos[right_arm_1] - arm_setpos)/arm_amp;
  thepos[3] = (double)cub_vel[right_arm_1]/arm_amp;
  thepos[4] = ((double)cub_pos[left_leg_1] - leg_setpos)/leg_amp;
  thepos[5] = (double)cub_vel[left_leg_1]/leg_amp;
  thepos[6] = ((double)cub_pos[right_leg_1] - leg_setpos)/leg_amp;
  thepos[7] = (double)cub_vel[right_leg_1]/leg_amp;
  
  CPGcontroller.setCurrentPosition(thepos);

  static char buffer_collision[50];

 //CPGcontroller.setDiscreteGoals(myIK.goals_rhythmic);
 CPGcontroller.setAmplitude(mu_crawl); 

  //We update the CPG
  CPGcontroller.integrate((double)(time_step)/1000.0,(double)(time_step)/1000.0);
  
  ///we get the state of the CPG
  double curr_state[4]; //current state of the CPG - x
  double curr_speed[4]; //current state of the CPG - y

  for (int i=0;i<4;i++)
    {
      curr_state[i] = CPGcontroller.getState(2*i)->getValue();
      curr_speed[i] = CPGcontroller.getState(2*i+1)->getValue();
    }
  
  double lelbow = 0.6*exp(-5.0*(curr_speed[0]+1.0)*(curr_speed[0]+1.0));
  double relbow = 0.6*exp(-5.0*(curr_speed[1]+1.0)*(curr_speed[1]+1.0));

  double lhip2 = - 0.2*exp(-5.0*(curr_speed[2]+1.0)*(curr_speed[2]+1.0));
  double rhip2 = 0.2*exp(-5.0*(curr_speed[3]+1.0)*(curr_speed[3]+1.0));

  //discrete system
  //it goes as left:  arm1 12 ,arm2 14, elbow 16 hip1 18
  // then      right: arm1 20 ,arm2 22, elbow 24 hip1 26
  cub_target_pos[left_arm_1] = arm_setpos + left_turn*arm_amp*curr_state[0];//arm_setpos offset 
  cub_target_pos[right_arm_1] = arm_setpos + right_turn*arm_amp*curr_state[1];
  cub_target_pos[left_leg_1] = leg_setpos + left_turn*leg_amp*curr_state[2];
  cub_target_pos[right_leg_1] = leg_setpos + right_turn*leg_amp*curr_state[3];
  



  cub_target_pos[left_elbow] = lelbow+CPGcontroller.getState(16)->getValue();
  cub_target_pos[right_elbow] = relbow+CPGcontroller.getState(24)->getValue();

  cub_target_pos[left_arm_2] = -0.5*lelbow+CPGcontroller.getState(14)->getValue();
  cub_target_pos[right_arm_2] = 0.5*relbow+CPGcontroller.getState(22)->getValue();

  cub_target_pos[left_leg_2] = lhip2-0.1;
  cub_target_pos[right_leg_2] = rhip2+0.1;

  cub_target_pos[left_forearm] = -M_PI/2.0;
  cub_target_pos[right_forearm] = M_PI/2.0;
  


  float orientation = 1.4;//-1.4: offset de base

  float torso, l_normal_elbow, l_normal_wrist; 
  torso=M_PI/2-orientation;
  //cout << "torso = " << torso <<"\n";
  l_normal_elbow = M_PI-torso-(cub_target_pos[left_arm_1]);
  //cout << "normal_elbow = " << normal_elbow <<"\n";
  l_normal_wrist = -cub_target_pos[left_elbow]+l_normal_elbow;
  //cout << "normal_wrist = " << normal_wrist << "\n";
  //cub_target_pos[left_wrist_2] = -l_normal_wrist;//+1.4;
  //cout << "wrist = " << cub_target_pos[left_wrist_2] <<"\n"; 

  cub_target_pos[left_wrist_1] = 0.2;
  cub_target_pos[right_wrist_1] = 0.2;

  cub_target_pos[left_arm_3] = -0.5;
  cub_target_pos[right_arm_3] = 0.5;

  float r_normal_elbow, r_normal_wrist; 
  r_normal_elbow = M_PI-torso-(cub_target_pos[right_arm_1]);
  r_normal_wrist = -cub_target_pos[right_elbow]+r_normal_elbow;
  //cub_target_pos[right_wrist_2] = r_normal_wrist;//-1.4;

  //float r_knee, l_knee;
  cub_target_pos[left_knee]=-1.87;//-(cub_target_pos[left_leg_1]+1.3);
  cub_target_pos[right_knee]=-1.87;//-(cub_target_pos[right_leg_1]+1.3);
 

  set_positions();

  //we record the values
  saveRealTrajectory();
  savePlannedTrajectory();
  saveSensorValues();
  savePosition();
  saveState();
  
  CPGcontroller.saveParameters();

  return time_step;

}


