/***************************************************************************
 *   Copyright (C) 20067 by Ludovic Righetti   *
 *   ludovic.righetti-at-epfl.ch   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifndef ICUBCRAWL_H
#define ICUBCRAWL_H

#define extension 0
#define flexion 1

#define CRAWL 0
#define REACH 1
#define CRAWL_REACH 2
#define NUMBER_OF_OBSTACLES 5

#include "iCubControl.h"

#include "Crawling_CPG.h"

#include "ACO.h"
#include <sstream>


class iCubCrawl : public iCubControl
{
 public:
  iCubCrawl(int time_step=1);
  ~iCubCrawl();

  virtual int run(int dt);
  virtual void reset();

  void init(double om_sw, double om_st);

  void saveState(const char *name=NULL);

  static void set_next_position(float next[3], float init[3]);

  float loop_time;
 
 private:

  double mu_crawl, mu_reach;

  Crawling_CPG CPGcontroller;
  
  double arm_setpos, leg_setpos;
  double arm_amp, leg_amp;

  int elbow; 
  int state;
  int time_collision;
  int initial;

  float next_position[3];

  int observer, feedback, gait;
  
  double current_time;

  fstream stateFile;
  fstream markFile;

  double left_turn ;
  double right_turn ;
  
  vector<int> path;
  int currNode;
  int previousNode;
  int goalNode;
  int pathCounter;
  int size;
  double findOrientation(int currNode, int nextNode);
  double previousOrientation;
  bool turn_right;
  bool turn_left;
  double newOrientation;
  int count1;
  string currOrientation;
  int time;
  
  float ob_translation_and_rotation[NUMBER_OF_OBSTACLES][7];
  float initial_ob_translation_and_rotation[NUMBER_OF_OBSTACLES][7];
  NodeRef obstacle;
  bool IsObstacle(int pos);
  vector<int> obstacles;
  
  bool IsPlanChanged();
  /*static float green_box_translation_and_rotation[7];
  static float blue_box_translation_and_rotation[7];
  static float yellow_box_translation_and_rotation[7];
  static float brown_box_translation_and_rotation[7];
  static float gray_box_translation_and_rotation[7];
  static float pink_box_translation_and_rotation[7];*/
  
 
};

#endif
